"""Import shorthand and command-line tool for Huber baths."""

from huber.driver import Bath


def command_line(args=None):
    """Command-line interface to the Huber bath."""
    import argparse
    import asyncio
    import json
    parser = argparse.ArgumentParser(description="Control a Huber bath "
                                     "from the command line.")
    parser.add_argument('ip', help="The bath IP address.")
    parser.add_argument('--set-setpoint', '-s', default=None, type=float,
                        help="Sets the bath temperature setpoint.")
    args = parser.parse_args(args)

    async def print_state():
        async with Bath(args.ip) as bath:
            if args.set_setpoint is not None:
                await bath.set_setpoint(args.set_setpoint)
            print(json.dumps(await bath.get(), indent=4, sort_keys=True))

    asyncio.run(print_state())


if __name__ == '__main__':
    command_line()



"""Mock interface to a Huber bath."""

import asyncio
import random

from huber.driver import Bath as realBath


class Bath(realBath):
    """Mock interface to a Huber bath."""

    def __init__(self, *args, **kwargs):
        """Init fixed variables."""
        self.temp_setpoint = 50
        self.pump_setpoint = 500
        self.on = False

    async def _connect(self):
        """Mock creating the TCP connection."""
        self.open = True

    def close(self):
        """Mock closing the TCP connection."""
        self.open = False

    async def get(self):
        """Return data structure randomly populated."""
        await asyncio.sleep(random.random() * 0.25)
        return {
            'on': self.on,  # Temperature control (+pump) active
            'temperature': {
                'bath': 23.49,                  # Internal (bath) temperature, °C
                'process': 22.71,               # Process temperature, °C
                'setpoint': self.temp_setpoint  # Temperature setpoint, °C
            },
            'pump': {
                'pressure': random.random() * 1000,    # Pump head pressure, mbar
                'speed': random.random() * 1000,       # Pump speed, rpm
                'setpoint': self.pump_setpoint  # Pump speed setpoint, rpm
            },
            'status': {
                'circulating': random.choice([False, True]),  # True if device is circulating
                'controlling': random.choice([False, True]),  # True if temp control is active
                'error': False,                        # True if an uncleared error is present
                'pumping': random.choice([False, True]),      # True if pump is on
                'warning': False,                      # True if an uncleared warning is present
            },
            'fill': random.random(),             # Oil level, [0, 1]
            'maintenance': random.random() * 365,  # Time until maintenance alarm, days
        }

    async def start(self):
        """Start the controller and pump."""
        await asyncio.sleep(random.random())
        self.on = True

    async def stop(self):
        """Stop the controller and pump."""
        await asyncio.sleep(random.random() * 0.25)
        self.on = False

    async def get_setpoint(self, *args, **kwargs):
        """Return setpoint."""
        return (await self.get())['temperature']['setpoint']

    async def set_setpoint(self, value):
        """Set the temperature setpoint of the bath, in C."""
        await asyncio.sleep(random.random() * 0.25)
        self.temp_setpoint = value

    async def get_bath_temperature(self):
        """Get the internal temperature of the bath, in C."""
        return (await self.get())['temperature']['bath']

    async def get_process_temperature(self):
        """Get the (optionally installed) process temperature, in C."""
        return (await self.get())['temperature']['process']

    async def get_pump_pressure(self):
        """Get the bath pump outlet pressure, in mbar."""
        return (await self.get())['pump']['pressure']

    async def get_pump_speed(self):
        """Get the bath pump speed, in RPM."""
        return (await self.get())['pump']['speed']

    async def set_pump_speed(self, value):
        """Set the bath pump speed, in RPM."""
        await asyncio.sleep(random.random() * 0.25)
        self.pump_setpoint = value

    async def get_fill_level(self):
        """Get the thermostat fluid fill level, in [0, 1]."""
        return (await self.get()).get('fill')

    async def get_next_maintenance(self):
        """Get the number of days until next maintenance alarm."""
        return (await self.get()).get('maintenance')

    async def get_status(self):
        """Get bath status indicators. Useful for triggering alerts."""
        return (await self.get())['status']

    async def get_error(self):
        """Get the most recent error, as a dictionary."""
        return (await self.get()).get('error')

    async def get_warning(self):
        """Get the most recent warning, as a dictionary."""
        return (await self.get()).get('warning')



"""Test the driver correctly initializes and returns mocked data."""
import random
from json import loads
from unittest import mock

import pytest

from huber import command_line
from huber.mock import Bath

fixed_random = random.random()
fixed_choice = random.choice([False, True])


@pytest.fixture
def huber_driver():
    """Confirm the driver correctly initializes."""
    return Bath('fake ip')


@pytest.fixture
def expected_data():
    """Return the mocked data."""
    with mock.patch('random.random', return_value=fixed_random), \
         mock.patch('random.choice', return_value=fixed_choice):
        return {
            'on': False,  # Temperature control (+pump) active
            'temperature': {
                'bath': 23.49,                  # Internal (bath) temperature, °C
                'process': 22.71,               # Process temperature, °C
                'setpoint': 50                  # Temperature setpoint, °C
            },
            'pump': {
                'pressure': random.random() * 1000,    # Pump head pressure, mbar
                'speed': random.random() * 1000,       # Pump speed, rpm
                'setpoint': 500                # Pump speed setpoint, rpm
            },
            'status': {
                'circulating': random.choice([False, True]),  # True if device is circulating
                'controlling': random.choice([False, True]),  # True if temp control is active
                'error': False,                        # True if an uncleared error is present
                'pumping': random.choice([False, True]),      # True if pump is on
                'warning': False,                      # True if an uncleared warning exists
            },
            'fill': random.random(),             # Oil level, [0, 1]
            'maintenance': random.random() * 365,  # Time until maintenance alarm, days
        }


@pytest.mark.asyncio
async def test_get_data(huber_driver, expected_data):
    """Confirm that the driver returns correct values on get() calls."""
    with mock.patch('random.random', return_value=fixed_random), \
         mock.patch('random.choice', return_value=fixed_choice):
        assert expected_data == await huber_driver.get()


@mock.patch('huber.Bath', Bath)
def test_driver_cli(capsys, expected_data):
    """Confirm the commandline interface works."""
    with mock.patch('random.random', return_value=fixed_random), \
         mock.patch('random.choice', return_value=fixed_choice):
        command_line(['fakeip'])
        captured = loads(capsys.readouterr().out)
        assert expected_data == captured


@mock.patch('huber.Bath', Bath)
def test_driver_cli_setpoint(capsys, expected_data):
    """Confirm setting a setpoint via the commandline interface works."""
    with mock.patch('random.random', return_value=fixed_random), \
         mock.patch('random.choice', return_value=fixed_choice):
        command_line(['fakeip', '--set-setpoint', '1.23'])
        captured = loads(capsys.readouterr().out)
        expected_data['temperature']['setpoint'] = 1.23
        assert expected_data == captured



"""Utilities to handle Huber's encodings."""
import csv
import os

root = os.path.normpath(os.path.dirname(__file__))
with open(os.path.join(root, 'faults.csv'), encoding='utf8') as in_file:
    reader = csv.reader(in_file)
    next(reader)
    faults = {int(row[0]): {
        'code': int(row[0]),
        'type': row[1],
        'condition': row[2] if len(row) >= 3 else None,
        'recovery': row[3] if len(row) == 4 else None
    } for row in reader if row[0]}

fields = {
    'on': {
        'address': 0x14,
        'writable': True,
        'format': 'b',
        'range': (0, 1)
    },
    'temperature': {
        'setpoint': {
            'address': 0x00,
            'writable': True,
            'format': 'f',
            'range': (-151, 327)
        },
        'bath': {
            'address': 0x01,
            'format': 'f',
            'range': (-151, 327)
        },
        'process': {
            'address': 0x07,
            'format': 'f',
            'range': (-151, 327)
        }
    },
    'pump': {
        'pressure': {
            'address': 0x03,
            'format': 'f',
            'range': (0, 320)
        },
        'speed': {
            'address': 0x26,
            'format': 'd',
            'range': (0, 32000)
        },
        'setpoint': {
            'address': 0x48,
            'writable': True,
            'format': 'd',
            'range': (1500, 4500)
        }
    },
    'fill': {
        'address': 0x0f,
        'format': '%',
        'range': (-0.001, 1)
    },
    'maintenance': {
        'address': 0x5c,
        'format': 'd'
    },
    'status': {
        'address': 0x0a,
        'format': 'list',
        'list': {
            0: 'controlling',
            1: 'circulating',
            4: 'pumping',
            8: 'error',
            9: 'warning'
        }
    },
    'error': {
        'address': 0x05,
        'format': 'fault',
        'writable': True
    },
    'warning': {
        'address': 0x06,
        'format': 'fault',
        'writable': True
    }
}


def int_to_hex(number, bits=16):
    """Convert int to a four-character hex string.

    This uses two's complement in encoding the integer, per Huber's
    manual.
    """
    return f'{(number + (1 << bits)) % (1 << bits):04X}'


def hex_to_int(hex_string, bits=16):
    """Convert a four-character hex string to an integer.

    This uses two's complement in decoding the hex string, per Huber's
    manual. This also manages the value '7fff' (32767), which is
    Huber's version of a command-not-implemented error.
    """
    number = int(hex_string, bits)
    if number == 32767:
        raise OSError("Command not enabled on this Huber model.")
    if number >> (bits - 1) == 1:
        number -= (1 << bits)
    return number


def parse(number, settings):
    """Parse a number according to formats from Huber's manual."""
    if number is None:
        return None
    format = settings['format']
    if format == 'b':
        return bool(number)
    if format == 'd':
        return number
    elif format == 'f':
        return number / 100.0
    elif format == '%':
        return number / 1000.0
    elif format == 'list':
        return {v: bool(number >> i & 1) for i, v in settings['list'].items()}
    elif format == 'fault':
        return faults[number] if number < 0 else None
    else:
        raise NotImplementedError(f'Number format "{format}" not supported.')


def get_field(key):
    """Search period-separated key searching on `fields`."""
    f = fields
    for k in key.split('.'):
        f = f[k]  # type: ignore
    return f


def set_nested(dictionary, keys, value):
    """Use period-separated keys to set nested dictionary value."""
    d = dictionary
    keys = keys.split('.')
    nodes, leaf = keys[:-1], keys[-1]
    for node in nodes:
        if node not in d:
            d[node] = {}
        d = d[node]
    d[leaf] = value
